-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "auxiliary"
M.relevantDevice = nil

local min = math.min
local max = math.max
local random = math.random

local rocketIndex = 1
local activeRockets = {}
local activeRocketCount = 0
local availableRockets = {}

local continuousRocketFire = false
local continuousRocketFireTimer = 0
local continuousRocketFireTime = 0

local function updateRocketWavefront(rocket, dt)
  rocket.timer = rocket.timer + dt
  if rocket.timer > 0.5 then
    rocket.timer = 0
    obj:breakBeam(rocket.wavefrontCid)
    rocket.update = nop
    rocket.state = "dead"
    rocket.isAlive = false
  end
end

local function updateRocketExploding(rocket, dt)
  obj:breakBeam(rocket.triggerCid)
  sounds.playSoundOnceAtNode("missileExplodeSound", (rocket.thrusterNode2), 15)
  for _,n in ipairs(rocket.fireCids) do
    local random = random(2) - 1
    obj:addParticleByNodes(rocket.nodeCid, n, 10 * random, 31, 1, 4)
  end
  electrics.values[rocket.motor] = nil
  rocket.update = updateRocketWavefront
  rocket.state = "wavefront"
end

local function updateRocketFlying(rocket, dt)
  local velocityVector = vec3(obj:getNodeVelocityVector(rocket.nodeCid))
  local velocity = velocityVector:length()
  local velocityError = rocket.targetVelocity - velocity
  local thrust = min(max(velocityError * 0.1, 0), 1)
  electrics.values[rocket.motor] = thrust
end

local function updateRocketFiring(rocket, dt)
  rocket.timer = rocket.timer + dt
  if rocket.timer > 0.2 then
    rocket.update = updateRocketFlying
    rocket.state = "flying"
    rocket.timer = 0

    for _,cid in ipairs(rocket.warheadCids) do
      obj:setNodeMass(cid, rocket.explosionNodeMass)
    end
  end
end

local function update(dt)
  for i = 1, activeRocketCount, 1 do
    activeRockets[i]:update(dt)
  end
end

local function fireRocket()
  for _ = 0, 1, 1 do
    if availableRockets[rocketIndex] then
      local rocket = availableRockets[rocketIndex]
      rocket.id = rocketIndex
      rocket.state = "firing"
      rocket.update = updateRocketFiring
      rocket.isAlive = true
      rocket.timer = 0
      rocket.hasFuel = rocket.fuel > 0
      electrics.values[rocket.motor] = 1
      sounds.playSoundOnceAtNode("missileLaunchSound", (rocket.thrusterNode2), 15)
      table.insert(activeRockets, rocket)
      activeRocketCount = #activeRockets
      rocketIndex = rocketIndex - 1
      beamstate.detachCouplers(rocket.couplerTag)
    end
  end

  gui.message(string.format("Rockets: %d", max(rocketIndex, 0)), 3, "rockets")
  continuousRocketFire = true
  continuousRocketFireTimer = continuousRocketFireTime
end

local function stopContinousRocketFire()
  continuousRocketFire = false
  continuousRocketFireTimer = 0
end

local function updateGFX(dt)
  if continuousRocketFire then
    continuousRocketFireTimer = continuousRocketFireTimer - dt
    if continuousRocketFireTimer <= 0 then
      fireRocket()
      continuousRocketFireTimer = continuousRocketFireTime
    end
  end

  local activeRockets2 = {}
  for _,v in ipairs(activeRockets) do
    if v.isAlive then
      v.fuel = max(v.fuel - dt, 0)
      if v.fuel <= 0 and v.hasFuel then
        v.update = nop
        electrics.values[v.motor] = nil
        v.hasFuel = false
      end

      if obj:beamIsBroken(v.collisionCid) and v.state ~= "wavefront" then
        v.update = updateRocketExploding
        v.state = "exploding"
      end
      table.insert(activeRockets2, v)

      if v.hasFuel and (v.id - rocketIndex - 1) < 4 then
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, -1, 37, 0.01, 4)
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, 1, 37, 0.01, 4)
      end
    end
  end

  activeRockets = activeRockets2
  activeRocketCount = #activeRockets
end

local function init(jbeamData)
  --dump(jbeamData)

  activeRockets = {}
  availableRockets = {}
  activeRocketCount = 0
  continuousRocketFire = false
  continuousRocketFireTimer = 0
  continuousRocketFireTime = jbeamData.continuousFireDelay or 0.3

  local rocketTargetVelocity = jbeamData.targetVelocity or 500
  local rocketFuel = jbeamData.fuel or 5
  local rocketExplosionNodeMass = jbeamData.explosionNodeMass or 30

  local rocketFireRefs = {}
  local rocketWarheadRefs = {}
  for k, node in pairs (v.data.nodes) do
    if node.fireNodeID then
      if not rocketFireRefs[node.fireNodeID] then
        rocketFireRefs[node.fireNodeID]= {}
      end
      table.insert(rocketFireRefs[node.fireNodeID], k)
    end
    if node.warheadNodeID then
      if not rocketWarheadRefs[node.warheadNodeID] then
        rocketWarheadRefs[node.warheadNodeID]= {}
      end
      table.insert(rocketWarheadRefs[node.warheadNodeID], k)
    end
  end

  local triggerRef = {}
  local collisionDetectorRef = {}
  local wavefrontDisableRef = {}
  for _, beam in pairs(v.data.beams) do
    if beam.weaponTriggerID then triggerRef[beam.weaponTriggerID] = beam.cid end
    if beam.collisionDetectorID then collisionDetectorRef[beam.collisionDetectorID] = beam.cid end
    if beam.wavefrontDisableID then wavefrontDisableRef[beam.wavefrontDisableID] = beam.cid end
  end

  for k,v in pairs(jbeamData) do
    if string.sub(k, 1, 6) == "rocket" then
      local rocketID = v.id
      local electricsPrefix = "rocket"..rocketID
      local rocket = {
        id = rocketID,
        state = "idle",
        nodeCid = v.centerNode,
        couplerTag = "rktrl_"..rocketID,
        motor = electricsPrefix.."_motor",
        triggerCid = triggerRef[rocketID],
        collisionCid = collisionDetectorRef[rocketID],
        wavefrontCid = wavefrontDisableRef[rocketID],
        fireCids = rocketFireRefs[rocketID],
        warheadCids = rocketWarheadRefs[rocketID],
        trailNodeFront = v.trailNodeFront,
        trailNodeBack = v.trailNodeBack,
        thrusterNode1 = v.thrusterNodeBack,
        thrusterNode2 = v.thrusterNodeFront,
        targetVelocity = v.targetVelocity or rocketTargetVelocity,
        fuel = v.fuel or rocketFuel,
        explosionNodeMass = v.explosionNodeMass or rocketExplosionNodeMass,
      }
      electrics.values[rocket.motor] = 0
      table.insert(availableRockets, rocket)
    end
  end

  table.sort(availableRockets,
    function(a,b)
      return a.id < b.id
    end)

  --dump(availableRockets)

  rocketIndex = #availableRockets

  math.randomseed(os.time())
end

M.init = init
M.updateGFX = updateGFX
M.update = update
M.deployWeapon = fireRocket
M.cancelWeaponCam = stopContinousRocketFire
M.toggleTargetMode = nop

return M