-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.outputPorts = {[1] = false}
M.deviceCategories = {engine = true}

local max = math.max
local min = math.min
local floor = math.floor

local function scaleFriction(device, friction)
  device.friction = device.friction * friction
end

local function scaleOutputTorque(device, state, maxReduction)
  device.outputTorqueState = max(device.outputTorqueState * state, maxReduction or 0)
end

local function disable(device)
  device.isDisabled = true
end

local function enable(device)
  device.isDisabled = false
end

local function setBoosterState(device, state)
  if device.hasBooster and device.boosterUsedDuration < device.boosterDuration then
    device.isBoosterActive = state
  end
end

local function setRawEngineThrottle(device, value)
  device.throttleSmoother:set(min(max(value, 0), 1))
end

local function updateEnergyStorageRatios(device)
  device.energyStorageRatios = {}
  for _,s in pairs(device.registeredEnergyStorages) do
    local storage = energyStorage.getStorage(s)
    if storage then
      if storage.storedEnergy > 0 then
        device.energyStorageRatios[storage.name] = 1 / device.storageWithEnergyCounter
      else
        device.energyStorageRatios[storage.name] = 0
      end
    end
  end
end

local function updateFuelUsage(device)
  if not device.energyStorage then
    return
  end

  local hasFuel = false
  local previousTankCount = device.storageWithEnergyCounter
  for _,s in pairs(device.registeredEnergyStorages) do
    local storage = energyStorage.getStorage(s)
    if storage then
      local previous = device.previousEnergyLevels[storage.name]
      storage.storedEnergy = max(storage.storedEnergy - (device.spentEnergy * device.energyStorageRatios[storage.name]), 0)
      if previous > 0 and storage.storedEnergy <= 0 then
        device.storageWithEnergyCounter = device.storageWithEnergyCounter - 1
      elseif previous <= 0 and storage.storedEnergy > 0 then
        device.storageWithEnergyCounter = device.storageWithEnergyCounter + 1
      end
      device.previousEnergyLevels[storage.name] = storage.storedEnergy
    end

    hasFuel = hasFuel or (storage and storage.storedEnergy > 0 or false)
  end
  if previousTankCount ~= device.storageWithEnergyCounter then
    device:updateEnergyStorageRatios()
  end
  device.spentEnergy = 0

  if not hasFuel and device.hasFuel then
    device:disable()
  elseif hasFuel and not device.hasFuel then
    device:enable()
  end

  device.hasFuel = hasFuel
end

local function updateGFX(device, dt)
  device:updateFuelUsage()

  if device.isBoosterActive then
    device.boosterUsedDuration = device.boosterUsedDuration + dt
    if device.boosterUsedDuration >= device.boosterDuration then
      device.isBoosterActive = false
    end
  end

  device.engineWorkPerUpdate = 0
  device.frictionLossPerUpdate = 0
  device.pumpingLossPerUpdate = 0
end

local function updateTorque(device, dt)
  device.throttle = min(max((electrics.values[device.electricsThrottleName] or 0) * (electrics.values[device.electricsThrottleFactorName] or 1), 0), 1)
  device.throttleSmoothed = device.throttleSmoother:getUncapped(device.throttle, dt) * (device.isDisabled and 0 or 1)
  device.thrust = device.maxThrust * device.throttleSmoothed + (device.isBoosterActive and device.boosterThrust or 0)
  obj:applyForce(device.thrustCid1, device.thrustCid2 , device.thrust)

  local usedFuel = device.specificFuelConsumption * device.thrust * dt / 1000 / 3600
  local thrustEnergy = 43.1 * 1000000 * usedFuel
  local exhaustRatio = min(max(floor((electrics.values.airspeed or 0) / device.maxExhaustVelocity * 100) * 0.01, 0), 1)
  device.spentEnergy = device.spentEnergy + thrustEnergy / device.burnEfficiencyTable[exhaustRatio]
end

local function selectUpdates(device)
  device.velocityUpdate = nop
  device.torqueUpdate = updateTorque
end

local function validate(device)
  return true
end

local function onBreak(device)
  device:disable()
end

local function registerStorage(device, storageName)
  local storage = energyStorage.getStorage(storageName)
  if storage and storage.storedEnergy > 0 then
    device.storageWithEnergyCounter = device.storageWithEnergyCounter + 1
    table.insert(device.registeredEnergyStorages, storageName)
    device:updateEnergyStorageRatios()
  end
  device.hasFuel = true
  device.previousEnergyLevels[storageName] = storage.storedEnergy
end

local function calculateInertia(device)
  device.cumulativeInertia = 0
  device.cumulativeGearRatio = 1
  device.maxCumulativeGearRatio = 1
end

local function new(jbeamData)
  local device = {
    deviceCategories = shallowcopy(M.deviceCategories),
    requiredExternalInertiaOutputs = shallowcopy(M.requiredExternalInertiaOutputs),
    outputPorts = shallowcopy(M.outputPorts),

    name = jbeamData.name,
    type = jbeamData.type,
    inputName = jbeamData.inputName,
    inputIndex = jbeamData.inputIndex,
    cumulativeInertia = 1,
    cumulativeGearRatio = 1,
    maxCumulativeGearRatio = 1,
    isPhysicallyDisconnected = false,
    isPropulsed = true,

    outputAV1 = 0,
    inputAV = 0,
    outputTorque1 = 0,
    isBroken = false,

    thrustCid1 = jbeamData.thrustNode1,
    thrustCid2 = jbeamData.thrustNode2,
    maxThrust = jbeamData.maxThrust or 0,
    maxExhaustVelocity = jbeamData.maxExhaustVelocity or 600,
    hasBooster = (jbeamData.boosterThrust or 0) > 0,
    boosterThrust = jbeamData.boosterThrust or 0,
    boosterDuration = jbeamData.boosterDuration or 0,
    isBoosterActive = false,
    boosterUsedDuration = 0,

    electricsThrottleName = jbeamData.electricsThrottleName or "throttle",
    electricsThrottleFactorName = jbeamData.electricsThrottleFactorName or "throttleFactor",
    throttle = 0,
    throttleSmoothed = 0,
    thrust = 0,
    specificFuelConsumption = jbeamData.specificFuelConsumption or 100,

    ignitionCoef = 1,
    isDisabled = false,

    throttleSmoother = newTemporalSmoothing(0.1, 0.09),

    spentEnergy = 0,
    storageWithEnergyCounter = 0,
    registeredEnergyStorages = {},
    previousEnergyLevels = {},
    hasFuel = true,

    onBreak = onBreak,
    validate = validate,
    calculateInertia = calculateInertia,
    updateGFX = updateGFX,
    setBoosterState = setBoosterState,
    scaleFriction = scaleFriction,
    scaleOutputTorque = scaleOutputTorque,
    disable = disable,
    enable = enable,
    updateFuelUsage = updateFuelUsage,
    updateEnergyStorageRatios = updateEnergyStorageRatios,
    registerStorage = registerStorage,

    setRawEngineThrottle = setRawEngineThrottle,

  }

  local tempBurnEfficiencyTable = nil
  if not jbeamData.burnEfficiency or type(jbeamData.burnEfficiency) == "number" then
    tempBurnEfficiencyTable = { {0, jbeamData.burnEfficiency or 1}, {1, jbeamData.burnEfficiency or 1}}
  elseif type(jbeamData.burnEfficiency) == "table" then
    tempBurnEfficiencyTable = deepcopy(jbeamData.burnEfficiency)
  end

  local copy = deepcopy(tempBurnEfficiencyTable)
  tempBurnEfficiencyTable = {}
  for k,v in pairs(copy) do
    if type(k) == "number" then
      table.insert(tempBurnEfficiencyTable, {v[1] * 100, v[2]})
    end
  end

  tempBurnEfficiencyTable = createCurve(tempBurnEfficiencyTable)
  device.burnEfficiencyTable = {}
  for k,v in pairs(tempBurnEfficiencyTable) do
    device.burnEfficiencyTable[k * 0.01] = v
  end

  device.requiredEnergyType = jbeamData.requiredEnergyType or "kerosene"
  device.energyStorage = jbeamData.energyStorage

  if device.thrustNode1 and device.thrustNode2 then
    local pos1 = vec3(v.data.nodes[device.thrustNode1].pos)
    local pos2 = vec3(v.data.nodes[device.thrustNode2].pos)
    local avgPos = (pos1 + pos2) / 2
    device.visualPosition = {x = avgPos.x, y = avgPos.y, z = avgPos.z}
  end

  --dump(jbeamData)

  device.breakTriggerBeam = jbeamData.breakTriggerBeam
  if device.breakTriggerBeam and device.breakTriggerBeam == "" then
    --get rid of the break beam if it's just an empty string (cancellation)
    device.breakTriggerBeam = nil
  end

  selectUpdates(device)

  return device
end

M.new = new

return M
